pub static APP_ID: &str = "com.belmoussaoui.Obfuscate";
pub static PKGDATADIR: &str = "/usr/share/obfuscate";
pub static PROFILE: &str = "";
pub static VERSION: &str = "0.0.10";
pub static GETTEXT_PACKAGE: &str = "obfuscate";
pub static LOCALEDIR: &str = "/usr/share/locale";
pub const RESOURCES_FILE: &str = concat!("/usr/share/obfuscate", "/obfuscate.gresource");
